DEFAULT_AZURE_API_KEY_NAME = "AZURE_API_KEY"
LITE_LLM_API_BASE_FIELD = "api_base"
LITE_LLM_API_KEY_FIELD = "api_key"
LITE_LLM_API_VERSION_FIELD = "api_version"
LITE_LLM_MODEL_FIELD = "model"
LITE_LLM_AZURE_AD_TOKEN = "azure_ad_token"
