// Copied from Chakra base design to make accessible via auto-completion
export const rasaSizes = {
  max: "max-content",
  min: "min-content",
  full: "100%",
};
export type RasaSizes = typeof rasaSizes;
