import{p as e,f as o}from"./flowDb-956e92f1-445b4048.js";import{f as t,g as a}from"./styles-c10674c1-6d90c832.js";import{u as i}from"./index-511b27e7.js";import"./graph-5f28327c.js";import"./layout-f169de61.js";import"./index-3862675e-adf68ef9.js";import"./clone-7300c31a.js";import"./edges-e0da2a9e-27d29830.js";import"./createText-2e5e7dd3-3ebaf8fc.js";import"./line-4404b723.js";import"./array-9f3ba611.js";import"./path-53f90ab3.js";import"./channel-03ccc28c.js";const M={parser:e,db:o,renderer:t,styles:a,init:r=>{r.flowchart||(r.flowchart={}),r.flowchart.arrowMarkerAbsolute=r.arrowMarkerAbsolute,i({flowchart:{arrowMarkerAbsolute:r.arrowMarkerAbsolute}}),t.setConf(r.flowchart),o.clear(),o.setGen("gen-2")}};export{M as diagram};
